import { Writable, Transform, PassThrough } from 'stream';
import fs from 'fs';
import path from 'path';
import mkdirp from 'mkdirp';
import dateFormat from './dateformat';

export class FileOutput extends Writable {
  constructor(config = {
    interval: 600,
    maxSize: 1024 * 1024,
    name: '%Y-%m-%d-%H-%M-%r-log.log',
    codec: null,
    dir: './logs/',
  }) {
    // If codec is specified, we'd be using object mode.
    super({
      objectMode: config.codec != null,
    });
    this.name = config.name || '%Y-%m-%d-%H-%M-%R-log.log';
    this.interval = config.interval || 600;
    this.maxSize = config.maxSize || 1024 * 1024;
    this.codec = config.codec;
    this.dir = config.dir || './logs/';

    // If the name is not fixed, we have to carry multiple file stream
    // at the single time. However, since the interval is fixed, we don't need
    // to care about the data per each file.
    // Name must be fixed if codec is not provided, though.
    this.isDynamic = typeof this.name === 'function';
    // Record currently ingested file size.
    this.currentSize = 0;
    // Create file swapper worker.
    mkdirp.sync(this.dir);
    this.refresher = setInterval(this.refresh.bind(this), 1000);
    this.refresh(true);
  }
  refresh(force, callback) {
    if (!(
      this.currentSize > this.maxSize ||
      (Date.now() - this.refreshedAt) > this.interval * 1000 ||
      force
    )) return;
    this.refreshedAt = Date.now();
    this.currentSize = 0;
    // Finish the ingestion, if any.
    if (this.isDynamic) {
      // TODO Add async.all?
      let remainingCount = 0;
      let innerCallback = () => {
        remainingCount--;
        if (remainingCount === 0 && callback) callback();
      };
      for (let key in this.files) {
        remainingCount++;
        this.finalizeFile(this.files[key], innerCallback);
      }
      this.files = {};
      if (remainingCount === 0 && callback) callback();
    } else if (this.file != null) {
      this.finalizeFile(this.file, callback);
      // Create new file
      this.file = null;
    } else {
      if (callback) callback();
    }
  }
  getFileName(nameTemplate) {
    let date = new Date(this.refreshedAt);
    // Format the name to fill in the details, such as date and random id.
    let name = dateFormat(date, nameTemplate);
    name = name.replace(/%r/g, () => Math.random() * 1e8 | 0);
    return path.resolve(this.dir, name);
  }
  createFile(nameTemplate) {
    // TODO Add mkdirp. This may not be possible since fs stream must be
    // initialized synchronously.
    let fileStream = fs.createWriteStream(
      this.getFileName(nameTemplate),
      { flags: 'a' });
    let stream = fileStream;
    // Before putting the data to file stream, we need to calculate how much
    // data we've used. We'll wrap a transform stream for that.
    let counterStream = new Transform({
      transform: (chunk, encoding, callback) => {
        if (chunk instanceof Buffer) {
          this.currentSize += chunk.length;
          callback(null, chunk);
        } else {
          let buffer = Buffer.from(chunk, encoding);
          this.currentSize += buffer.length;
          callback(null, buffer);
        }
      },
    });
    // If codec is specified, create a passthrough stream and call codec with
    // it, then pipe the returned value of codec to file stream.
    if (this.codec) {
      stream = new PassThrough({ objectMode: true });
      this.codec(stream).pipe(counterStream).pipe(fileStream);
    } else {
      stream = counterStream;
      counterStream.pipe(fileStream);
    }
    return {
      stream,
      fileStream,
    };
  }
  finalizeFile(file, callback) {
    // Finalize the stream and end the data ingestion, then send to AWS,
    // then delete the file.
    file.stream.end();
    file.fileStream.once('close', callback);
  }
  _write(chunk, encoding, callback) {
    // We need to honor the backpressure, thus we're forwarding the callback.
    this.refresh();
    if (this.isDynamic) {
      let name = this.name(chunk);
      let file = this.files[name];
      if (file == null) {
        file = this.files[name] = this.createFile(name);
      }
      file.stream.write(chunk, encoding, callback);
    } else {
      if (this.file == null) {
        this.file = this.createFile(this.name);
      }
      this.file.stream.write(chunk, encoding, callback);
    }
  }
  _final(callback) {
    clearInterval(this.refresher);
    this.refresh(true, callback);
  }
}

export default function createFileOutput(config) {
  return new FileOutput(config);
}

module.exports = createFileOutput;
createFileOutput.FileOutput = FileOutput;
